/**
 * ================================================================
 * TopologicalSorting01.java: Manually compute topological sort ...
 * 
 * Written by: Mark Austin                                July 2018
 * ================================================================
 */

package demo;

import java.util.*;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;

import java.lang.Boolean;

import java.util.List;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;

public class TopologicalSorting01  {

     static class Node {
       public final String name;
       public final HashSet<Edge> inEdges;
       public final HashSet<Edge> outEdges;
       public Node(String name) {
          this.name = name;
          inEdges   = new HashSet<Edge>();
          outEdges  = new HashSet<Edge>();
       }

       public Node addEdge(Node node){
          Edge e = new Edge(this, node);
          outEdges.add(e);
          node.inEdges.add(e);
          return this;
       }

       @Override
       public String toString() {
          StringBuffer buf = new StringBuffer();

          buf.append( String.format("Node: %s ...\n", name));
          buf.append( String.format("--- no  in edges = %d ...\n", inEdges.size() ));
          buf.append( String.format("--- no out edges = %d ...\n", outEdges.size() ));

          return buf.toString();
       }
     }

     static class Edge {
        public final Node from;
        public final Node   to;
        public Edge(Node from, Node to) {
           this.from = from;
           this.to = to;
        }

        @Override
        public boolean equals(Object obj) {
           Edge e = (Edge)obj;
           return e.from == from && e.to == to;
        }
     }

     public static void main(String[] args) {

        System.out.println("Run TopologicalSorting.main() ...");
        System.out.println("====================================== ...");

        // Assemble nodes and edge dependencies ...

        System.out.println("--- Create nodes ...");

        Node two    = new Node("2");
        Node three  = new Node("3");
        Node five   = new Node("5");
        Node seven  = new Node("7");
        Node eight  = new Node("8");
        Node nine   = new Node("9");
        Node ten    = new Node("10");
        Node eleven = new Node("11");

        System.out.println("--- Add edge dependencies ...");

        seven.addEdge(eleven).addEdge(eight);
        five.addEdge(eleven);
        three.addEdge(eight).addEdge(ten);
        eleven.addEdge(two).addEdge(nine).addEdge(ten);
        eight.addEdge(nine).addEdge(ten);

        System.out.println("--- Print details of nodes ...\n");

        System.out.println( two.toString () );
        System.out.println( three.toString () );
        System.out.println( five.toString () );
        System.out.println( seven.toString () );
        System.out.println( eight.toString () );
        System.out.println( nine.toString () );
        System.out.println( ten.toString () );
        System.out.println( eleven.toString () );

        System.out.println("--- Create array of nodes ...");

        Node[] allNodes = {seven, five, three, eleven, eight, two, nine, ten };

        // L <- Empty list that will contain the sorted elements

        ArrayList<Node> L = new ArrayList<Node>();

        //S <- Set of all nodes with no incoming edges

        System.out.println("--- Create set of nodes with no incoming edges ...");

        HashSet<Node> S = new HashSet<Node>(); 
        for(Node n : allNodes){
           if(n.inEdges.size() == 0){
              S.add(n);
           }
        }

        System.out.println("--- Manually compute dependencies ...");

        // while S is non-empty do

        while(!S.isEmpty()) {

            // remove a node n from S

            Node n = S.iterator().next();
            S.remove(n);

            // insert n into L

            L.add(n);

           // for each node m with an edge e from n to m do

           for(Iterator<Edge> it = n.outEdges.iterator();it.hasNext();){

               // remove edge e from the graph

               Edge e = it.next();
               Node m = e.to;
               it.remove();//Remove edge from n
               m.inEdges.remove(e);//Remove edge from m

               // if m has no other incoming edges then insert m into S

               if(m.inEdges.isEmpty()){
                  S.add(m);
               }
           }
        }

        // Check to see if all edges are removed

        System.out.println("--- Check to see if all edges are removed ...");

        boolean cycle = false;
        for(Node n : allNodes){
           if(!n.inEdges.isEmpty()){
              cycle = true;
              break;
           }
        }

        // Print result ....

        if( cycle == true ){
            System.out.println("--- Cycle present, topological sort not possible");
        } else {
            System.out.println("--- Topological Sort: "+ Arrays.toString(L.toArray() ));
        }

        System.out.println("====================================== ...");
        System.out.println("Done !! ...");
   }
}
