/*
 * =====================================================================
 * (C) Copyright 2018-2018, by Timofey Chudakov and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * =====================================================================
 */

package demo.jgrapht.alg.clique;

import org.jgrapht.*;
import org.jgrapht.alg.clique.*;
import org.jgrapht.graph.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

/**
 * Tests for the {@link ChordalGraphMaxCliqueFinder}
 *
 * @author Timofey Chudakov
 */

public class ChordalGraphMaxCliqueFinderTest {
    /**
     * Tests maximum clique finding on an empty graph.
     */
    @Test
    public void testGetMaximumClique1() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testGetMaximumClique1() ...\n");

        Graph<Integer, DefaultEdge> graph = new DefaultUndirectedGraph<>(DefaultEdge.class);
        Set<Integer> clique = new ChordalGraphMaxCliqueFinder<>(graph).getClique();

        System.out.printf("--- Results: %s \n", clique.toString() );

        assertNotNull(clique);
        assertEquals(0, clique.size());

        System.out.printf("--- Leave testGetMaximumClique1() ...\n");
    }

    /**
     * Tests maximum clique finding on a chordal graph
     */
    @Test
    public void testGetMaximumClique2() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testGetMaximumClique2() ...\n");

        Graph<Integer, DefaultEdge> graph = new DefaultUndirectedGraph<>(DefaultEdge.class);
        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 3, 4);
        Graphs.addEdgeWithVertices(graph, 3, 5);
        Graphs.addEdgeWithVertices(graph, 3, 6);
        Graphs.addEdgeWithVertices(graph, 4, 5);
        Graphs.addEdgeWithVertices(graph, 4, 6);
        Graphs.addEdgeWithVertices(graph, 5, 6);

        System.out.printf("--- Test graph: %s \n", graph.toString() );

        Set<Integer> clique = new ChordalGraphMaxCliqueFinder<>(graph).getClique();

        System.out.printf("--- Results: %s \n", clique.toString() );

        assertNotNull(clique);
        assertEquals(4, clique.size());
        assertIsClique(graph, clique);

        System.out.printf("--- Leave testGetMaximumClique2() ...\n");
    }

    /**
     * Tests maximum clique finding on a non-chordal graph
     */

    @Test
    public void testGetMaximumClique3() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testGetMaximumClique3() ...\n");

        Graph<Integer, DefaultEdge> graph = new DefaultUndirectedGraph<>(DefaultEdge.class);
        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 2, 3);
        Graphs.addEdgeWithVertices(graph, 3, 4);
        Graphs.addEdgeWithVertices(graph, 1, 4);

        System.out.printf("--- Test graph: %s \n", graph.toString() );

        Set<Integer> clique = new ChordalGraphMaxCliqueFinder<>(graph).getClique();

        if(clique != null)
           System.out.printf("--- Results: %s \n", clique.toString() );
        else
           System.out.printf("--- Results: clique = null \n");

        System.out.printf("--- Leave testGetMaximumClique3() ...\n");
    }

    /**
     * Tests maximum clique finding on a pseudograph
     */

    @Test
    public void testGetMaximumClique4() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testGetMaximumClique4() ...\n");

        Graph<Integer, DefaultEdge> graph = new Pseudograph<>(DefaultEdge.class);
        Graphs.addEdgeWithVertices(graph, 1, 1);
        Graphs.addEdgeWithVertices(graph, 1, 1);
        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 2, 2);
        Graphs.addEdgeWithVertices(graph, 2, 2);

        System.out.printf("--- Test graph: %s \n", graph.toString() );

        Set<Integer> clique = new ChordalGraphMaxCliqueFinder<>(graph).getClique();

        System.out.printf("--- Results: %s \n", clique.toString() );

        assertNotNull(clique);
        assertEquals(2, clique.size());
        assertIsClique(graph, clique);

        System.out.printf("--- Leave testGetMaximumClique4() ...\n");
    }

    /**
     * Tests maximum clique finding on a non-chordal graph
     */

    @Test
    public void testGetMaximumClique5() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testGetMaximumClique5() ...\n");
        System.out.printf("--- Network of triangle cliques ...\n");

        Graph<Integer, DefaultEdge> graph = new DefaultUndirectedGraph<>(DefaultEdge.class);

        // Four-node clique of  ...

        Graphs.addEdgeWithVertices(graph, 1, 2);
        Graphs.addEdgeWithVertices(graph, 1, 3);
        Graphs.addEdgeWithVertices(graph, 1, 7);
        Graphs.addEdgeWithVertices(graph, 2, 3);
        Graphs.addEdgeWithVertices(graph, 2, 7);
        Graphs.addEdgeWithVertices(graph, 3, 7);

        // Three-node clique of  ...

        Graphs.addEdgeWithVertices(graph, 5, 6);
        Graphs.addEdgeWithVertices(graph, 5, 4);
        Graphs.addEdgeWithVertices(graph, 6, 4);

        // Connect cliques with line of edges ...

        Graphs.addEdgeWithVertices(graph, 1, 8);
        Graphs.addEdgeWithVertices(graph, 8, 9);
        Graphs.addEdgeWithVertices(graph, 9, 5);

        System.out.printf("--- Test graph: %s \n", graph.toString() );

        Set<Integer> clique = new ChordalGraphMaxCliqueFinder<>(graph).getClique();

        if(clique != null)
           System.out.printf("--- Results: %s \n", clique.toString() );
        else
           System.out.printf("--- Results: clique = null \n");

        System.out.printf("--- Leave testGetMaximumClique5() ...\n");
    }

    /**
     * Checks whether every two vertices from {@code set} are adjacent.
     *
     * @param graph the tested graph.
     * @param set the tested set of vertices.
     * @param <V> the graph vertex type.
     * @param <E> the graph edge type.
     */

    private <V, E> void assertIsClique(Graph<V, E> graph, Set<V> set) {
        ArrayList<V> vertices = new ArrayList<>(set);
        for (int i = 0; i < vertices.size(); i++) {
            for (int j = 0; j < i; j++) {
                assertTrue(graph.containsEdge(vertices.get(i), vertices.get(j)));
            }
        }
    }
}
