/*
 * =========================================================================
 * (C) Copyright 2016-2018, by Vera-Licona Research Group and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * =========================================================================
 */

package demo.jgrapht.alg.shortestpath;

import org.jgrapht.*;
import org.jgrapht.graph.*;
import org.jgrapht.alg.shortestpath.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

/**
 * Test cases for the AllDirectedPaths algorithm.
 *
 * @author Andrew Gainer-Dewar, Google LLC
 **/

public class AllDirectedPathsTest {
    private static final String I1 = "I1";
    private static final String I2 = "I2";
    private static final String A = "A";
    private static final String B = "B";
    private static final String C = "C";
    private static final String D = "D";
    private static final String E = "E";
    private static final String F = "F";
    private static final String O1 = "O1";
    private static final String O2 = "O2";

    @Test
    public void testSmallExampleGraph() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testSmallExampleGraph) ...\n");
        System.out.printf("--- ================================================= ...\n");

        System.out.printf("--- Toy graph: %s ...\n", toyGraph().toString() );

        AllDirectedPaths<String, DefaultEdge> pathFindingAlg = new AllDirectedPaths<>(toyGraph());

        Set<String> sources = new HashSet<>();
        sources.add(I1);
        sources.add(I2);

        Set<String> targets = new HashSet<>();
        targets.add(O1);
        targets.add(O2);

        System.out.printf("--- Set (sources): %s ...\n", sources.toString() );
        System.out.printf("--- Set (targets): %s ...\n", targets.toString() );

        System.out.printf("--- All paths analysis ...\n");

        List<GraphPath<String, DefaultEdge>> allPaths = pathFindingAlg.getAllPaths(sources, targets, true, null);

        System.out.printf("--- No paths = %d ...\n", allPaths.size() );

        for (int i = 1; i <= allPaths.size(); i = i + 1 ) {
           System.out.printf("--- Path %2d: %s ...\n", i, allPaths.get(i-1).toString() );
        }

        assertEquals("Toy network should have correct number of simple paths", 7, allPaths.size());

        System.out.printf("--- ================================================= ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testSmallExampleGraph) ...\n");
    }

    @Test
    public void testTrivialPaths() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testTrivialPaths() ...\n");
        System.out.printf("--- ============================================= ...\n");

        System.out.printf("--- Toy graph: %s ...\n", toyGraph().toString() );

        // Verify fix for http://github.com/jgrapht/jgrapht/issues/234.

        AllDirectedPaths<String, DefaultEdge> pathFindingAlg = new AllDirectedPaths<>(toyGraph());

        Set<String> sources = new HashSet<>();
        sources.add(I1);

        Set<String> targets = new HashSet<>();
        targets.add(I1);
        targets.add(A);

        System.out.printf("--- Set (sources): %s ...\n", sources.toString() );
        System.out.printf("--- Set (targets): %s ...\n", targets.toString() );

        System.out.printf("--- All paths analysis ...\n");

        List<GraphPath<String, DefaultEdge>> allPaths = pathFindingAlg.getAllPaths(sources, targets, true, 1);

        System.out.printf("---   No paths = %d ...\n", allPaths.size() );
        for (int i = 1; i <= allPaths.size(); i = i + 1 ) {
           System.out.printf("--- Path %2d: %s ...\n", i, allPaths.get(i-1).toString() );
        }

        assertEquals(
            "Toy network should have correct number of trivial simple paths", 2, allPaths.size());
        assertEquals(Arrays.asList(I1), allPaths.get(0).getVertexList());
        assertEquals(Arrays.asList(I1, A), allPaths.get(1).getVertexList());

        System.out.printf("--- ============================================= ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testTrivialPaths() ...\n");
    }

    @Test
    public void testLengthOnePaths() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testLengthOnePaths() ...\n");
        System.out.printf("--- ============================================== ...\n");

        // Verify fix for http://github.com/jgrapht/jgrapht/issues/441.

        DefaultDirectedGraph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);
        graph.addVertex("A");
        graph.addVertex("B");
        graph.addEdge("B", "A");

        System.out.printf("--- Graph: %s ...\n", graph.toString() );

        AllDirectedPaths<String, DefaultEdge> all = new AllDirectedPaths<>(graph);
        List<GraphPath<String, DefaultEdge>> allPaths =
            all.getAllPaths(graph.vertexSet(), graph.vertexSet(), true, graph.edgeSet().size());

        System.out.printf("--- No paths = %d ...\n", allPaths.size() );
        for (int i = 1; i <= allPaths.size(); i = i + 1 ) {
           System.out.printf("--- Path %2d: %s ...\n", i, allPaths.get(i-1).toString() );
        }

        assertEquals(3, allPaths.size());
        assertEquals(Arrays.asList("A"), allPaths.get(0).getVertexList());
        assertEquals(Arrays.asList("B"), allPaths.get(1).getVertexList());
        assertEquals(Arrays.asList("B", "A"), allPaths.get(2).getVertexList());

        System.out.printf("--- ============================================== ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testLengthOnePaths() ...\n");
    }

    @Test
    public void testPathWeights() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testPathWeights() ...\n");
        System.out.printf("--- ============================================= ...\n");

        // Verify fix for https://github.com/jgrapht/jgrapht/issues/617.

	SimpleDirectedWeightedGraph<String, DefaultWeightedEdge> graph =
	    new SimpleDirectedWeightedGraph<>(DefaultWeightedEdge.class);

        graph.addVertex("A");
	graph.addVertex("B");
	graph.addVertex("C");
	graph.addVertex("D");
	graph.addVertex("E");
	graph.addVertex("F");

	graph.setEdgeWeight(graph.addEdge("A", "B"), 1.2);
	graph.setEdgeWeight(graph.addEdge("A", "C"), 0);
	graph.setEdgeWeight(graph.addEdge("A", "D"), -1);
	graph.setEdgeWeight(graph.addEdge("B", "C"), 2);
	graph.setEdgeWeight(graph.addEdge("B", "D"), 1);
	graph.setEdgeWeight(graph.addEdge("C", "D"), 0.5);
	graph.setEdgeWeight(graph.addEdge("D", "E"), 1.0);
	graph.setEdgeWeight(graph.addEdge("E", "F"), 1.0);
	graph.setEdgeWeight(graph.addEdge("F", "C"), 1.0);

        System.out.printf("--- Graph: %s ...\n", graph.toString() );

        // Set search parameters ...

        String source = "A";
        String target = "F";
        int maxPathLength = 4;

        System.out.printf("--- Source node: %s ...\n", source);
        System.out.printf("--- Target node: %s ...\n", target);
        System.out.printf("--- Max path length = %d ...\n", maxPathLength );

	AllDirectedPaths<String, DefaultWeightedEdge> all = new AllDirectedPaths<>(graph);
        List<GraphPath<String, DefaultWeightedEdge>> allPaths = all.getAllPaths( source, target, true, maxPathLength);

        System.out.printf("--- \n");
        System.out.printf("--- No paths = %d ...\n", allPaths.size() );
        for (int i = 1; i <= allPaths.size(); i = i + 1 ) {
           System.out.printf("--- Path %2d: %s ...\n", i, allPaths.get(i-1).toString() );
        }

        System.out.printf("--- \n");
        System.out.printf("--- Sort paths by weight ...\n");

	allPaths.sort(Comparator.comparing(GraphPath::getWeight));

        for (int i = 1; i <= allPaths.size(); i = i + 1 ) {
           double pathWeight = allPaths.get(i-1).getWeight();
           System.out.printf("--- Path %2d: %5.2f: %s ...\n", i, pathWeight, allPaths.get(i-1).toString() );
        }

        System.out.printf("--- ============================================= ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testPathWeights() ...\n");
    }

    @Test
    public void testCycleBehavior() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testCycleBehavior() ...\n");
        System.out.printf("--- ============================================== ...\n");

        Graph<String, DefaultEdge> toyGraph = toyGraph();
        toyGraph.addEdge(D, A);

        AllDirectedPaths<String, DefaultEdge> pathFindingAlg = new AllDirectedPaths<>(toyGraph);

        Set<String> sources = new HashSet<>();
        sources.add(I1);
        sources.add(I2);

        Set<String> targets = new HashSet<>();
        targets.add(O1);
        targets.add(O2);

        List<GraphPath<String, DefaultEdge>> allPathsWithoutCycle =
            pathFindingAlg.getAllPaths(sources, targets, true, 8);

        List<GraphPath<String, DefaultEdge>> allPathsWithCycle =
            pathFindingAlg.getAllPaths(sources, targets, false, 8);

        System.out.printf("---   All paths without cycle: %s ...\n", allPathsWithoutCycle.toString() );
        System.out.printf("---   All simplepaths: %s ...\n", allPathsWithCycle.toString() );

        assertEquals( "Toy network with cycle should have correct number of paths with cycle", 13,
            allPathsWithCycle.size());
        assertEquals( "Toy network with cycle should have correct number of simple paths", 7,
            allPathsWithoutCycle.size());

        System.out.printf("--- ============================================== ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testCycleBehavior() ...\n");
    }

    @Test
    public void testMustBoundIfNonSimplePaths() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllDirectedPathsTest.testMustBoundIfNonSimplePaths() ...\n");
        System.out.printf("--- ========================================================== ...\n");

        // Goofy hack to test for an exception

        AllDirectedPaths<String, DefaultEdge> pathFindingAlg = new AllDirectedPaths<>(toyGraph());

        Set<String> sources = Collections.singleton(I1);
        Set<String> targets = Collections.singleton(O1);

        try {
            pathFindingAlg.getAllPaths(sources, targets, false, null);
            fail("Expected an IllegalArgumentException");
        } catch (IllegalArgumentException e) {
            // This is the expected outcome, so the test passes
        }

        System.out.printf("--- ========================================================== ...\n");
        System.out.printf("--- Leave AllDirectedPathsTest.testMustBoundIfNonSimplePaths() ...\n");
    }

    private static Graph<String, DefaultEdge> toyGraph() {

        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);
        graph.addVertex(I1);
        graph.addVertex(I2);
        graph.addVertex(A);
        graph.addVertex(B);
        graph.addVertex(C);
        graph.addVertex(D);
        graph.addVertex(E);
        graph.addVertex(F);
        graph.addVertex(O1);
        graph.addVertex(O2);

        graph.addEdge(I1, A);
        graph.addEdge(I1, B);

        graph.addEdge(I2, B);
        graph.addEdge(I2, C);

        graph.addEdge(A, B);
        graph.addEdge(A, D);
        graph.addEdge(A, E);

        graph.addEdge(B, E);

        graph.addEdge(C, B);
        graph.addEdge(C, F);

        graph.addEdge(D, E);

        graph.addEdge(E, O1);

        graph.addEdge(F, O2);

        return graph;
    }
}
