/*
 * ======================================================================
 * (C) Copyright 2006-2018, by John V Sichi and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * ======================================================================
 */

package demo.jgrapht.alg.shortestpath;

import org.jgrapht.*;
import org.jgrapht.alg.interfaces.ShortestPathAlgorithm.*;
import org.jgrapht.alg.shortestpath.*;
import org.jgrapht.graph.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

/**
 * .
 *
 * @author John V. Sichi
 */
public class BellmanFordShortestPathTest extends ShortestPathTestCase
{
    // ~ Methods ----------------------------------------------------------------

    @Test
    public void testUndirected() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter BellmanFordShortestPathTest.testUndirected() ...\n");
        System.out.printf("--- ================================================== ...\n");

        System.out.printf("--- Create graph ...\n");

        Graph<String, DefaultWeightedEdge> g = create();

        System.out.printf("--- Graph: %s ...\n", g.toString() );

        System.out.printf("--- \n");
        System.out.printf("--- Extract tree of paths from V1 ...\n");

        SingleSourcePaths<String, DefaultWeightedEdge> tree = new BellmanFordShortestPath<>(g).getPaths(V1);

        System.out.printf("--- Path(V1-->V1): %s ...\n", tree.getPath(V1).toString() );
        System.out.printf("--- Path(V1-->V2): %s ...\n", tree.getPath(V2).toString() );
        System.out.printf("--- Path(V1-->V3): %s ...\n", tree.getPath(V3).toString() );
        System.out.printf("--- Path(V1-->V4): %s ...\n", tree.getPath(V4).toString() );
        System.out.printf("--- Path(V1-->V5): %s ...\n", tree.getPath(V5).toString() );

        System.out.printf("--- \n");
        System.out.printf("--- Extract tree of paths from V3 ...\n");

        tree = new BellmanFordShortestPath<>(g).getPaths(V3);

        System.out.printf("--- Path(V3-->V1): %s ...\n", tree.getPath(V1).toString() );
        System.out.printf("--- Path(V3-->V2): %s ...\n", tree.getPath(V2).toString() );
        System.out.printf("--- Path(V3-->V3): %s ...\n", tree.getPath(V3).toString() );
        System.out.printf("--- Path(V3-->V4): %s ...\n", tree.getPath(V4).toString() );
        System.out.printf("--- Path(V3-->V5): %s ...\n", tree.getPath(V5).toString() );

        System.out.printf("--- ================================================== ...\n");
        System.out.printf("--- Leave BellmanFordShortestPathTest.testUndirected() ...\n");
    }

    @Override
    protected List<DefaultWeightedEdge> findPathBetween(
        Graph<String, DefaultWeightedEdge> g, String src, String dest)
    {
        return new BellmanFordShortestPath<>(g).getPaths(src).getPath(dest).getEdgeList();
    }

    @Test
    public void testWithNegativeEdges()
    {
        Graph<String, DefaultWeightedEdge> g = createWithBias(true);

        List<DefaultWeightedEdge> path;

        path = findPathBetween(g, V1, V4);
        assertEquals(Arrays.asList(e13, e34), path);

        path = findPathBetween(g, V1, V5);
        assertEquals(Arrays.asList(e15), path);
    }

    @Test
    public void testNoPath()
    {
        DirectedWeightedPseudograph<String, DefaultWeightedEdge> g =
            new DirectedWeightedPseudograph<>(DefaultWeightedEdge.class);
        g.addVertex("a");
        g.addVertex("b");

        BellmanFordShortestPath<String, DefaultWeightedEdge> alg = new BellmanFordShortestPath<>(g);
        SingleSourcePaths<String, DefaultWeightedEdge> paths = alg.getPaths("a");
        assertEquals(paths.getWeight("b"), Double.POSITIVE_INFINITY, 0);
        assertNull(paths.getPath("b"));
    }

    @Test
    public void testWikipediaExampleBellmanFord() {

        System.out.printf("--- Enter BellmanFordShortestPathTest.testWikipediaBellmanFord() ...\n");
        System.out.printf("--- ============================================================ ...\n");

        System.out.printf("--- Assemble directed weighed pseudograph ...\n");

        DirectedWeightedPseudograph<String, DefaultWeightedEdge> g =
            new DirectedWeightedPseudograph<>(DefaultWeightedEdge.class);
        g.addVertex("w");
        g.addVertex("y");
        g.addVertex("x");
        g.addVertex("z");
        g.addVertex("s");
        g.setEdgeWeight(g.addEdge("w", "z"), 2);
        g.setEdgeWeight(g.addEdge("y", "w"), 4);
        g.setEdgeWeight(g.addEdge("x", "w"), 6);
        g.setEdgeWeight(g.addEdge("x", "y"), 3);
        g.setEdgeWeight(g.addEdge("z", "x"), -7);
        g.setEdgeWeight(g.addEdge("y", "z"), 5);
        g.setEdgeWeight(g.addEdge("z", "y"), -3);
        g.setEdgeWeight(g.addEdge("s", "w"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "y"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "x"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "z"), 0.0);

        System.out.printf("--- Graph: %s ...\n", g.toString() );

        System.out.printf("--- Compute shortest paths ...\n");

        BellmanFordShortestPath<String, DefaultWeightedEdge> alg = new BellmanFordShortestPath<>(g);

        System.out.printf("--- Print summary of paths from \"s\" ...\n");

        SingleSourcePaths<String, DefaultWeightedEdge> tree = alg.getPaths("s");

        System.out.printf("--- Path(s-->s): %s ...\n", tree.getPath("s").toString() );
        System.out.printf("--- Path(s-->w): %s ...\n", tree.getPath("w").toString() );
        System.out.printf("--- Path(s-->x): %s ...\n", tree.getPath("x").toString() );
        System.out.printf("--- Path(s-->y): %s ...\n", tree.getPath("y").toString() );
        System.out.printf("--- Path(s-->z): %s ...\n", tree.getPath("z").toString() );

        System.out.printf("--- ============================================================ ...\n");
        System.out.printf("--- Leave BellmanFordShortestPathTest.testWikipediaBellmanFord() ...\n");
    }

    @Test
    public void testNegativeCycleDetection() {
        DirectedWeightedPseudograph<String, DefaultWeightedEdge> g =
            new DirectedWeightedPseudograph<>(DefaultWeightedEdge.class);
        g.addVertex("w");
        g.addVertex("y");
        g.addVertex("x");
        g.addVertex("z");
        g.addVertex("s");
        g.setEdgeWeight(g.addEdge("w", "z"), 2);
        g.setEdgeWeight(g.addEdge("y", "w"), 4);
        g.setEdgeWeight(g.addEdge("x", "w"), 6);
        g.setEdgeWeight(g.addEdge("x", "y"), 3);
        g.setEdgeWeight(g.addEdge("z", "x"), -7);
        g.setEdgeWeight(g.addEdge("y", "z"), 3);
        g.setEdgeWeight(g.addEdge("z", "y"), -3);
        g.setEdgeWeight(g.addEdge("s", "w"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "y"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "x"), 0.0);
        g.setEdgeWeight(g.addEdge("s", "z"), 0.0);

        try {
            new BellmanFordShortestPath<>(g).getPaths("s");
            fail("Negative-weight cycle not detected");
        } catch (RuntimeException e) {
            assertEquals("Graph contains a negative-weight cycle", e.getMessage());
        }
    }

    @Test
    public void testNegativeEdgeUndirectedGraph() {
        WeightedPseudograph<String, DefaultWeightedEdge> g =
            new WeightedPseudograph<>(DefaultWeightedEdge.class);
        g.addVertex("w");
        g.addVertex("y");
        g.addVertex("x");
        g.setEdgeWeight(g.addEdge("w", "y"), 1);
        g.setEdgeWeight(g.addEdge("y", "x"), 1);
        g.setEdgeWeight(g.addEdge("y", "x"), -1);
        try {
            new BellmanFordShortestPath<>(g).getPaths("w");
            fail("Negative-weight cycle not detected");
        } catch (RuntimeException e) {
            assertEquals("Graph contains a negative-weight cycle", e.getMessage());
        }
    }

}

// End BellmanFordShortestPathTest.java
