/*
 * =======================================================================
 * (C) Copyright 2016-2017, by Dimitrios Michail and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * =======================================================================
 */

package demo.jgrapht.io;

import org.jgrapht.*;
import org.jgrapht.graph.*;
import org.jgrapht.io.*;
import org.junit.*;

import java.io.*;
import java.util.*;

import static org.junit.Assert.*;

/**
 * 2nd part of tests for DOTImporter. See also {@link DOTImporter1Test}.
 */

public class DOTImporter2Test {
    private static final String NL = "\n";

    @Test
    public void testDOT1() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testDOT1() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph { a -> b -> c; }";
        // @formatter:on

        // Assemble graph from dot input string ...

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        DirectedPseudograph<String, DefaultEdge> graph = new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);

        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        importer.importGraph(graph, new StringReader(input));

        // Manually assemble expected graph ...

        DirectedPseudograph<String, DefaultEdge> expected =
                       new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a", "b", "c"));
        expected.addEdge("a", "b");
        expected.addEdge("b", "c");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        // Make assertion ...

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testDOT1() ...\n");
    }

    @Test
    public void testDOT2() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testDOT2() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  subgraph cluster0 { " + NL +
                       "   a0 -> a1 -> a2 -> a3;" + NL +
                       "  }" + NL +
                       "  subgraph cluster1 { " + NL +
                       "   b0 -> b1 -> b2 -> b3;" + NL +
                       "  }" + NL +
                       "  start -> a0;" + NL +
                       "  start -> b0;" + NL +
                       "  a1 -> b3;" + NL +
                       "  b2 -> a3;" + NL +
                       "  a3 -> a0;" + NL +
                       "  a3 -> end;" + NL +
                       "  b3 -> end;" + NL +
                       "}";
        // @formatter:on

        // Assemble graph from dot input string ...

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        // Manually assemble expected graph ...

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(
            expected,
            Arrays.asList("a0", "a1", "a2", "a3", "b0", "b1", "b2", "b3", "start", "end"));
        expected.addEdge("a0", "a1");
        expected.addEdge("a1", "a2");
        expected.addEdge("a2", "a3");
        expected.addEdge("b0", "b1");
        expected.addEdge("b1", "b2");
        expected.addEdge("b2", "b3");
        expected.addEdge("start", "a0");
        expected.addEdge("start", "b0");
        expected.addEdge("a1", "b3");
        expected.addEdge("b2", "a3");
        expected.addEdge("a3", "a0");
        expected.addEdge("a3", "end");
        expected.addEdge("b3", "end");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        // Make assertion ...

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("======================================= ...\n");
        System.out.printf("Leave testDOT2() ...\n");
    }

    @Test
    public void testDOT3() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testDOT3() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  subgraph { " + NL +
                       "   a0 -> a1;" + NL +
                       "  }" + "->" +
                       "  subgraph { " + NL +
                       "   b0 -> b1;" + NL +
                       "  }" + NL +
                       "}";
        // @formatter:on

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );

        // Assemble graph from dot input string ...

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        // Manually assemble expected graph ...

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a1", "b0", "b1"));
        expected.addEdge("a0", "a1");
        expected.addEdge("b0", "b1");
        expected.addEdge("a0", "b0");
        expected.addEdge("a0", "b1");
        expected.addEdge("a1", "b0");
        expected.addEdge("a1", "b1");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        // Make assertion ...

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testDOT3() ...\n");
    }

    @Test
    public void testDOT4() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testDOT4() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  subgraph { " + NL +
                       "   a0 -> a1;" + NL +
                       "   subgraph { " + NL +
                       "     a00 -> a11" + NL +
                       "   }" + NL +
                       "  }" + "->" +
                       "  subgraph { " + NL +
                       "   b0 -> b1;" + NL +
                       "   subgraph { " + NL +
                       "     b00 -> b11" + NL +
                       "   }" + NL +
                       "  }" + NL +
                       "}";
        // @formatter:on

        // Assemble graph from dot input string ...

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        // Manually assemble expected graph ...

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(
            expected, Arrays.asList("a0", "a1", "a00", "a11", "b0", "b1", "b00", "b11"));
        expected.addEdge("a0", "a1");
        expected.addEdge("a00", "a11");
        expected.addEdge("b0", "b1");
        expected.addEdge("b00", "b11");
        expected.addEdge("a0", "b0");
        expected.addEdge("a0", "b1");
        expected.addEdge("a0", "b00");
        expected.addEdge("a0", "b11");
        expected.addEdge("a1", "b0");
        expected.addEdge("a1", "b1");
        expected.addEdge("a1", "b00");
        expected.addEdge("a1", "b11");
        expected.addEdge("a00", "b0");
        expected.addEdge("a00", "b1");
        expected.addEdge("a00", "b00");
        expected.addEdge("a00", "b11");
        expected.addEdge("a11", "b0");
        expected.addEdge("a11", "b1");
        expected.addEdge("a11", "b00");
        expected.addEdge("a11", "b11");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        // Make assertion ...

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testDOT4() ...\n");
    }

    @Test
    public void testNodeSubgraphEdges() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testNodeSubgraphEdges() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  a0 -> { a00 -> a11 } -> b0 -> { b00 -> b11 }" + NL +
                       "}";
        // @formatter:on

        // Assemble graph from dot input string ...

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        // Manually assemble expected graph ...

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a00", "a11", "b0", "b00", "b11"));
        expected.addEdge("a00", "a11");
        expected.addEdge("b00", "b11");
        expected.addEdge("a0", "a00");
        expected.addEdge("a0", "a11");
        expected.addEdge("a00", "b0");
        expected.addEdge("a11", "b0");
        expected.addEdge("b0", "b00");
        expected.addEdge("b0", "b11");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        // Make assertion ...

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testNodeSubgraphEdges() ...\n");
    }

    @Test
    public void testNodeSubgraphEdgesUndirected() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testNodeSubgraphEdgesUndirected() ...\n");
        System.out.printf("--- ======================================= ...\n");


        // @formatter:off
        String input = "digraph G {" + NL +
                       "  a0 -> { a00 -- a11 } -> b0 -- { b00 -- b11 }" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a00", "a11", "b0", "b00", "b11"));
        expected.addEdge("a00", "a11");
        expected.addEdge("b00", "b11");
        expected.addEdge("a0", "a00");
        expected.addEdge("a0", "a11");
        expected.addEdge("a00", "b0");
        expected.addEdge("a11", "b0");
        expected.addEdge("b0", "b00");
        expected.addEdge("b0", "b11");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testNodeSubgraphEdgesUndirected() ...\n");
    }

    @Test
    public void testNestedSubgraphs() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testNestedSubgraphs() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  a0 -> { { { { a00 -> a11 } } } } -> b0 -> { { b00 -> b11 } }" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a00", "a11", "b0", "b00", "b11"));
        expected.addEdge("a00", "a11");
        expected.addEdge("b00", "b11");
        expected.addEdge("a0", "a00");
        expected.addEdge("a0", "a11");
        expected.addEdge("a00", "b0");
        expected.addEdge("a11", "b0");
        expected.addEdge("b0", "b00");
        expected.addEdge("b0", "b11");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testNestedSubgraphs() ...\n");
    }

    @Test
    public void testEdgeAttributes() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testEdgeAttributes() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  edge [weight=5.0];" + NL +
                       "  a0 -> a1;" + NL +
                       "  subgraph {" + NL +
                       "    edge [weight=2.0];" + NL +
                       "    a2 -> a3;" + NL +
                       "  };" + NL +
                       "  a4 -> a5 [weight=15.0];" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        Map<DefaultEdge, Map<String, Attribute>> edgeAttributes = new HashMap<>();
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> {
            DefaultEdge e = new DefaultEdge();
            edgeAttributes.put(e, a);
            return e;
        };
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a1", "a2", "a3", "a4", "a5"));
        expected.addEdge("a0", "a1");
        expected.addEdge("a2", "a3");
        expected.addEdge("a4", "a5");
        assertEquals(expected.toString(), graph.toString());

        for (DefaultEdge e : edgeAttributes.keySet()) {
            String f = graph.getEdgeSource(e);
            String t = graph.getEdgeSource(e);
            Map<String, Attribute> attrs = edgeAttributes.get(e);
            if (f.equals("a0") && t.equals("a1")) {
                assertEquals("5.0", attrs.get("weight"));
            } else if (f.equals("a2") && t.equals("a3")) {
                assertEquals("5.0", attrs.get("weight"));
            } else if (f.equals("a4") && t.equals("a5")) {
                assertEquals("15.0", attrs.get("weight"));
            }
        }

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testEdgeAttributes() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void testComments() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testComments() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G { // ignore" + NL +
                       "  /* ignore */ a0 -> a1; /* ignore */" + NL +
                       "# ignore" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a1"));
        expected.addEdge("a0", "a1");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        assertEquals(expected.toString(), graph.toString());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testComments() ...\n");
    }

    @Test
    public void testNodeAttributes() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testNodeAttributes() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  node [color=gray];" + NL +
                       "  a0 -> a1;" + NL +
                       "  subgraph {" + NL +
                       "    node [color=black];" + NL +
                       "    a2 -> a3;" + NL +
                       "  };" + NL +
                       "  a4 [color=white];" + NL +                       
                       "  a4 -> a5;" + NL +
                       "}";
        // @formatter:on

        Map<String, Map<String, Attribute>> vertexAttributes = new HashMap<>();
        VertexProvider<String> vp = (label, attrs) -> {
            vertexAttributes.put(label, attrs);
            return label;
        };
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        DirectedPseudograph<String, DefaultEdge> expected =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        Graphs.addAllVertices(expected, Arrays.asList("a0", "a1", "a2", "a3", "a4", "a5"));
        expected.addEdge("a0", "a1");
        expected.addEdge("a2", "a3");
        expected.addEdge("a4", "a5");

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Expected String: %s ...\n", expected.toString() );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        assertEquals(expected.toString(), graph.toString());

        // assertEquals("gray", vertexAttributes.get("a0").get("color").getValue());
        // assertEquals("gray", vertexAttributes.get("a1").get("color").getValue());
        // assertEquals("black", vertexAttributes.get("a2").get("color").getValue());
        // assertEquals("black", vertexAttributes.get("a3").get("color").getValue());
        // assertEquals("white", vertexAttributes.get("a4").get("color").getValue());
        // assertEquals("gray", vertexAttributes.get("a5").get("color").getValue());

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testNodeAttributes() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void testError() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testError() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  edge [weight=5.0];" + NL +
                       "  a0 -> a1;" + NL +
                       "  subgraph {" + NL +
                       "    edge [weight=2.0];" + NL +
                       "    a2 -> a3;" + NL +
                       "  };" + NL +
                       "  a4 -> a5 [weight=15.0];" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        Map<DefaultEdge, Map<String, Attribute>> edgeAttributes = new HashMap<>();
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> {
            DefaultEdge e = new DefaultEdge();
            if (f.equals("a2") && t.equals("a3")) {
                throw new Error("Error");
            }
            edgeAttributes.put(e, a);
            return e;
        };
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph =
            new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        try {
            importer.importGraph(graph, new StringReader(input));
            fail("No!");
        } catch (Error e) {
            assertEquals("Error", e.getMessage());
        }

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testError() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void testLoopError() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testLoopError() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  a0 -> a0;" + NL +
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();
        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedMultigraph<String, DefaultEdge> graph = new DirectedMultigraph<String, DefaultEdge>(DefaultEdge.class);
        try {
            importer.importGraph(graph, new StringReader(input));
            fail("No!");
        } catch (ImportException e) {
        }

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testLoopError() ...\n");
    }

    @Test
    public void testExampleDOT1() throws ImportException {

        System.out.printf("--- \n");
        System.out.printf("--- Enter testExampleDOT1() ...\n");
        System.out.printf("--- ======================================= ...\n");

        // @formatter:off
        String input = "digraph G {" + NL +
                       "  subgraph cluster0 {" + NL +
                       "    node [style=filled,color=white];" + NL+
                       "    style=filled;" + NL+
                       "    color=lightgrey;" + NL+
                       "    a0->a1->a2->a3;" + NL+
                       "    label=\"process #1\";" + NL+
                       "  }"+NL+
                       "  subgraph cluster1 {" + NL+
                       "    node [style=filled];"+NL+
                       "    b0->b1->b2->b3;" + NL+
                       "    label=\"process #2\";" + NL+
                       "    color=blue"+NL+
                       "  }"+NL+
                       "  start -> a0;"+NL+
                       "  start -> b0;"+NL+
                       "  a1 -> b3;"+NL+
                       "  b2 -> a3;"+NL+
                       "  a3 -> a0;"+NL+
                       "  a3 -> end;"+NL+
                       "  b3 -> end;"+NL+
                       "  start [shape=Mdiamond];"+NL+
                       "  end [shape=Msquare];" + NL+ 
                       "}";
        // @formatter:on

        VertexProvider<String> vp = (a, b) -> a;
        EdgeProvider<String, DefaultEdge> ep = (f, t, l, a) -> new DefaultEdge();

        GraphImporter<String, DefaultEdge> importer = new DOTImporter<String, DefaultEdge>(vp, ep);
        DirectedPseudograph<String, DefaultEdge> graph = new DirectedPseudograph<String, DefaultEdge>(DefaultEdge.class);
        importer.importGraph(graph, new StringReader(input));

        // Print strings ...

        System.out.printf("--- \n");
        System.out.printf("--- Input String: %s ...\n", input );
        System.out.printf("--- \n");
        System.out.printf("--- Graph String: %s ...\n", graph.toString() );

        System.out.printf("--- ======================================= ...\n");
        System.out.printf("--- Leave testExampleDOT1() ...\n");
        System.out.printf("--- \n");
    }
}
