/*
 * =========================================================================================
 * (C) Copyright 2016-2017, by Dimitrios Michail and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 *
 * -----------------------------------------------------------------------------------------
 * Notes: GRAPH --> Undirected Graph.
 *        DIGRAPH --> Directed Graph.
 * =========================================================================================
 */

package org.jgrapht.io;

import org.antlr.v4.runtime.*;
import org.antlr.v4.runtime.misc.*;
import org.apache.commons.lang3.*;
import org.apache.commons.lang3.text.translate.*;
import org.jgrapht.*;

import parser.antlr.dot.*;

import java.io.*;
import org.jgrapht.io.*;
import java.util.*;

/**
 * =========================================================================================
 * Import a graph from a DOT file.
 *
 * <p>
 * For a description of the format see <a href="http://en.wikipedia.org/wiki/DOT_language">
 * http://en.wikipedia.org/wiki/DOT_language</a> and
 * <a href="http://www.graphviz.org/doc/info/lang.html">
 * http://www.graphviz.org/doc/info/lang.html</a>
 * </p>
 *
 * @author Dimitrios Michail
 *
 * @param <V> the graph vertex type
 * @param <E> the graph edge type
 * =========================================================================================
 */

public class DOTImporter<V, E> extends AbstractBaseImporter<V, E> implements GraphImporter<V, E> {
    boolean DEBUG = true;

    /**
     * Default key used in the graph updater (if provided) for the graph ID.
     */
    public static final String DEFAULT_GRAPH_ID_KEY = "ID";

    // identifier unescape rule

    private static final CharSequenceTranslator UNESCAPE_ID = new AggregateTranslator(
        new LookupTranslator(
            new String[][] { { "\\\\", "\\" }, { "\\\"", "\"" }, { "\\'", "'" }, { "\\", "" } }));

    /**
     * Constructs a new importer.
     *
     * @param vertexProvider used to create vertices
     * @param edgeProvider used to create edges
     */

    public DOTImporter(VertexProvider<V> vertexProvider, EdgeProvider<V, E> edgeProvider) {
        this(vertexProvider, edgeProvider, null);
    }

    /**
     * Constructs a new importer.
     *
     * @param vertexProvider used to create vertices
     * @param edgeProvider used to create edges
     * @param vertexUpdater used to further update vertices
     */

    public DOTImporter(
        VertexProvider<V> vertexProvider, EdgeProvider<V, E> edgeProvider,
        ComponentUpdater<V> vertexUpdater) {
        this(vertexProvider, edgeProvider, vertexUpdater, null);
    }

    /**
     * Constructs a new importer.
     *
     * @param vertexProvider used to create vertices
     * @param edgeProvider used to create edges
     * @param vertexUpdater used to further update vertices
     * @param graphUpdater used to update graph attributes, like the graph identifier
     */

    public DOTImporter(
        VertexProvider<V> vertexProvider, EdgeProvider<V, E> edgeProvider,
        ComponentUpdater<V> vertexUpdater, ComponentUpdater<Graph<V, E>> graphUpdater) {
        super(vertexProvider, edgeProvider, (vertexUpdater != null) ? vertexUpdater : (c, a) -> {
        }, (graphUpdater != null) ? graphUpdater : (c, a) -> {
        });
    }

    /**
     * {@inheritDoc}
     */

    @Override
    public void importGraph(Graph<V, E> g, Reader in) throws ImportException {

        if(DEBUG == true) {
           System.out.printf("--- \n");
           System.out.printf("--- Enter jgrapht.io.DOTImporter.importGraph() ...\n");
        }

        try {
            /**
             * Create lexer with unbuffered input stream and use a token factory which copies
             * characters from the input stream into the text of the tokens.
             */

            if(DEBUG == true) {
               System.out.printf("---   Create DOT lexer ...\n");
            }

            DOTLexer lexer = new DOTLexer(new UnbufferedCharStream(in));
            lexer.setTokenFactory(new CommonTokenFactory(true));
            lexer.removeErrorListeners();
            ThrowingErrorListener errorListener = new ThrowingErrorListener();
            lexer.addErrorListener(errorListener);

            // Create parser with unbuffered token stream.

            if(DEBUG == true) {
               System.out.printf("---   Create parser with unbuffered token stream ...\n");
            }

            DOTParser parser = new DOTParser( new UnbufferedTokenStream<>(lexer) );
            parser.removeErrorListeners();
            parser.addErrorListener(errorListener);

            //  Disable parse tree building and attach listener.

            if(DEBUG == true) {
               System.out.printf("---   Disable parse tree building and attach listener ...\n");
            }

            parser.setBuildParseTree(false);

            if(DEBUG == true) {
               System.out.printf("---   Create GraphDOTListener ...\n");
            }

            CreateGraphDOTListener listener = new CreateGraphDOTListener(g);
            parser.addParseListener( listener );

            // Parse graph ...

            if(DEBUG == true) {
               System.out.printf("---   Goto parser.graph() ...\n");
            }

            parser.graph();

        } catch (ParseCancellationException | IllegalArgumentException e) {
            throw new ImportException("Failed to import DOT graph: " + e.getMessage(), e);
        }

        if(DEBUG == true) {
           System.out.printf("--- Leave DOTImporter.importGraph() ...\n");
        }
    }

    /*
     * Common error listener for both lexer and parser which throws an exception.
     */

    private class ThrowingErrorListener extends BaseErrorListener {

        @Override
        public void syntaxError(
            Recognizer<?, ?> recognizer, Object offendingSymbol, int line, int charPositionInLine,
            String msg, RecognitionException e)
            throws ParseCancellationException
        {
            throw new ParseCancellationException(
                "line " + line + ":" + charPositionInLine + " " + msg);
        }
    }

    /*
     * ======================================================================
     * Listen on parser events and construct the graph.
     * The listener is strongly dependent on the grammar.
     * ======================================================================
     */

    private class CreateGraphDOTListener extends DOTBaseListener {

        // Graph to update

        private Graph<V, E> graph;
        private Map<String, V> vertices;

        // Stacks to maintain scope and state

        private Deque<SubgraphScope> subgraphScopes;
        private Deque<State> stack;

        public CreateGraphDOTListener(Graph<V, E> graph) {
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener() ... \n");

            this.graph = graph;
            this.vertices       = new HashMap<>();
            this.stack          = new ArrayDeque<>();
            this.subgraphScopes = new ArrayDeque<>();

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener() ... \n");
        }

        @Override
        public void enterGraph(DOTParser.GraphContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterGraph() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // Process graph state ...

            stack.push(new State());
            subgraphScopes.push(new SubgraphScope());

            /*
             * Validate graph for directed or undirected. We do not validate for strict on purpose,
             * but let the user decide the behavior by using the appropriate graph class.
             */

            if (ctx.DIGRAPH() != null && !graph.getType().isDirected()) {
                throw new IllegalArgumentException("Provided graph is not directed");
            }
            if (ctx.GRAPH() != null && !graph.getType().isUndirected()) {
                throw new IllegalArgumentException("Provided graph is not undirected");
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterGraph() ... \n");
        }

        @Override
        public void exitGraph(DOTParser.GraphContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitGraph() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                System.out.printf("---     ERROR: stack.isEmpty() --> true ... \n");
                return;
            }
            subgraphScopes.pop();
            stack.pop();

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitGraph() ... \n");
        }

        public void enterGraphHeader(DOTParser.GraphContext ctx) {}
        public void exitGraphHeader(DOTParser.GraphContext ctx) {}

        public void enterId(DOTParser.GraphContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterId() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // add partial state

            stack.push(new State());

            System.out.printf("---     Exit DOTImporter.CreateGraphDOTListener.enterId() ... \n");
        }

        public void exitId(DOTParser.GraphContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitId() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty()) {
                return;
            }

            // read graph id

            State s = stack.pop();
            State idPartial = s.children.peekFirst();

            if (idPartial != null) {
                try {
                    // notify graph updater
                    graphUpdater.update( graph,
                    Collections.singletonMap( DEFAULT_GRAPH_ID_KEY, DefaultAttribute.createAttribute(idPartial.getId())));
                } catch (Exception e) {
                    throw new IllegalArgumentException("Graph update failed: " + e.getMessage(), e);
                }
            }

            // Add as child of parent

            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitId() ... \n");
        }

        public void enterAttr_stmt( DOTParser.Attr_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterAttr_stmt() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            stack.push(new State());    // add partial state

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterAttr_stmt() ... \n");
        }

        public void exitAttr_stmt( DOTParser.Attr_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitAttr_stmt() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                return;
            }

            // Read attributes

            State s     = stack.pop();
            State child = s.children.peekFirst();
            if (child != null && child.attrs != null) {
                Map<String, Attribute> attrs = child.attrs;

                // Update current scope

                SubgraphScope scope = subgraphScopes.element();

                if (ctx.NODE() != null) {
                    scope.nodeAttrs.putAll(attrs);
                } else if (ctx.EDGE() != null) {
                    scope.edgeAttrs.putAll(attrs);
                } else if (ctx.GRAPH() != null) {
                    scope.graphAttrs.putAll(attrs);
                }
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitAttr_stmt() ... \n");
        }

        public void enterA_list( DOTParser.A_listContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterA_list() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // add partial state
            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterA_list() ... \n");

        }

        public void exitA_list( DOTParser.A_listContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitA_list() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty()) {
                return;
            }

            // union children attributes

            State s = stack.pop();
            for (State child : s.children) {
                if (child.attrs != null) {
                    s.putAll(child.attrs);
                }
            }

            // add as child of parent
            s.children.clear();
            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitA_list() ... \n");
        }

        public void enterAttr_list( DOTParser.Attr_listContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterAttr_list() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // add partial state
            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterAttr_list() ... \n");
        }

        public void exitAttr_list( DOTParser.Attr_listContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterAttr_list() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty()) {
                return;
            }

            // Collect attributes in map

            State s = stack.pop();
            Iterator<State> it = s.children.iterator();
            while (it.hasNext()) {
                State child = it.next();
                if (child.ids != null && child.ids.size() == 1) {
                    s.put(child.ids.get(0), null);
                } else if (child.ids != null && child.ids.size() >= 2) {
                    s.put(child.ids.get(0), DefaultAttribute.createAttribute(child.ids.get(1)));
                }
                it.remove();
            }

            // Add as child of parent

            s.children.clear();
            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterAttr_list() ... \n");
        }

        public void enterEdge_stmt( DOTParser.Edge_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterEdge_stmt() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // add partial state
            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterEdge_stmt() ... \n");
        }

        public void exitEdge_stmt( DOTParser.Edge_stmtContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitEdge_stmt() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                return;
            }

            State s = stack.pop();

            // Add edges to model ...

            System.out.printf("---     Iterate over %d children states within current state ... \n", s.children.size() );

            for (int i = 1; i < s.children.size(); i = i + 1 ) {
                System.out.printf("--- Start loop %d ...\n", i );

                State prev = s.children.get(i-1);
                State  cur = s.children.get(i);

                // Connect edges between adjacent states ...

                List<V> prevVertices = prev.getVertices();
                List<V>  curVertices =  cur.getVertices();

                // Read attributes

                Map<String, Attribute> attrs = Collections.emptyMap();
                if (cur != null) {
                    attrs = cur.attrs;
                }

                // Connect edges between adjacent states ...

                for (V sourceVertex : prev.getVertices()) {
                for (V targetVertex : cur.getVertices())   {

                    // Process default and extra attributes

                    Map<String, Attribute> edgeAttrs = new HashMap<>(subgraphScopes.element().edgeAttrs);
                    if (attrs != null) {
                       edgeAttrs.putAll(attrs);
                    }

                    try {
                        String edgeLabel = null;
                        if (edgeAttrs.containsKey("label")) {
                            edgeLabel = edgeAttrs.get("label").toString();
                        }

                        E e = edgeProvider.buildEdge(sourceVertex, targetVertex, edgeLabel, edgeAttrs);

                        System.out.printf("---     Run graph.addEdge(): %s --> %s ... \n", sourceVertex, targetVertex );

                        graph.addEdge(sourceVertex, targetVertex, e);

                        System.out.printf("---     Edge added ... \n");

                    } catch (Exception e) {
                        throw new IllegalArgumentException( "Edge creation failed: " + e.getMessage(), e);
                    }
                }
                }
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitEdge_stmt() ... \n");
        }

        // =====================================================================
        // Pair Statements are used in parsing of HTML ...
        // =====================================================================

        public void exitIdentifierPairStatement( DOTParser.IdContext ctx ) {

            System.out.printf("---     Exit DOTImporter.CreateGraphDOTListener.exitIdentifiersPairStatement() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                return;
            }

            // Read key value pair

            State s = stack.pop();
            State idPairChild = s.children.peekFirst();
            if (idPairChild == null) {
                return;
            }
            String key   = idPairChild.ids.get(0);
            String value = idPairChild.ids.get(1);

            // Update attributes in current scope

            SubgraphScope scope = subgraphScopes.element();
            scope.graphAttrs.put(key, DefaultAttribute.createAttribute(value));
            if (subgraphScopes.size() == 1) {
                try {
                    graphUpdater.update( graph,
                        Collections.singletonMap(key, DefaultAttribute.createAttribute(value)));
                } catch (Exception e) {
                    throw new IllegalArgumentException("Graph update failed: " + e.getMessage(), e);
                }
            }
        }

        public void enterNode_stmt(DOTParser.Node_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterNode_stmt() ... \n");

            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterNode_stmt() ... \n");
        }

        public void exitNode_stmt(DOTParser.Node_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitNode_stmt() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
               System.out.printf("---     stack.isEmpty()          --> %s ...\n",
                                 Boolean.toString( stack.isEmpty() ) );
               System.out.printf("---     subgraphScopes.isEmpty() --> %s ...\n",
                                 Boolean.toString( subgraphScopes.isEmpty() ) );
               System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitNode_stmt() ... \n");
               return;
            }

            // Read node id

            State s = stack.pop();
            Iterator<State> it = s.children.iterator();
            if (it.hasNext() == false) {
                return;
            }

            State nodeIdPartialState = it.next();
            String nodeId = nodeIdPartialState.getId();

            System.out.printf("---     nodeId = %s ... \n", nodeId );

            // Read attributes

            System.out.printf("---     Read attributes ... \n");

            Map<String, Attribute> attrs = Collections.emptyMap();
            if (it.hasNext() == true) {
                attrs = it.next().attrs;
            }

            // Create or update vertex

            V v = vertices.get(nodeId);
            if (v == null) {

                System.out.printf("---     Create new vertex ... \n");

                // Find default attributes, append extra attributes ...

                SubgraphScope scope = subgraphScopes.element();
                Map<String, Attribute> defaultAttrs = new HashMap<>(scope.nodeAttrs);

                // Create vertex ...

                try {
                    v = vertexProvider.buildVertex(nodeId, defaultAttrs);
                } catch (Exception e) {
                    throw new IllegalArgumentException( "Vertex creation failed: " + e.getMessage(), e);
                }

                // Add vertex to graph ...

                System.out.printf("---     Run graph.addVertex(v) ... \n");

                graph.addVertex(v);

                // Save vertex to node hashmap ...

                System.out.printf("---     Add v to HashMap ... \n");

                vertices.put(nodeId, v);

                // Add vertex to scope ...

                System.out.printf("---     Add v to scope ... \n");

                if ( scope != null || v != null ) {
                   scope.addVertex(v);
                } else {
                   System.out.printf("--- WARNING: scope or vertex v is null ...\n");
                }

            } else {
                System.out.printf("---     Update vertex: goto vertexUpdater.update(...) ...\n");
                if ( v != null || attrs != null ) {
                   vertexUpdater.update(v, attrs);
                } else {
                   System.out.printf("---     WARNING: Args to vertexUpdater are null ...\n");
                }
            }

            // Add vertex to stack of states ...

            System.out.printf("---     Add v to state ... \n");

            s.addVertex(v);

            // Add as child of parent

            s.children.clear();
            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitNode_stmt() ... \n");
        }

        // -----------------------------------------------------------
        // NOTE: New ANTLR parser does not call these two methods ...
        // -----------------------------------------------------------

        public void enterNodeStatementNoAttributes( DOTParser.Node_stmtContext ctx) {
            // add partial state
            stack.push(new State());
        }

        public void exitNodeStatementNoAttributes( DOTParser.Node_stmtContext ctx) {

            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitNodeStatementNoAttribute() ... \n");

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                return;
            }

            // read node id
            State s = stack.pop();
            Iterator<State> it = s.children.iterator();
            if (!it.hasNext()) {
                return;
            }
            State nodeIdPartial = it.next();
            String nodeId = nodeIdPartial.getId();

            // create or update vertex
            V v = vertices.get(nodeId);
            if (v == null) {
                SubgraphScope scope = subgraphScopes.element();
                // find default attributes
                Map<String, Attribute> defaultAttrs = new HashMap<>(scope.nodeAttrs);
                try {
                    v = vertexProvider.buildVertex(nodeId, defaultAttrs);
                } catch (Exception e) {
                    throw new IllegalArgumentException(
                        "Vertex creation failed: " + e.getMessage(), e);
                }
                graph.addVertex(v);
                vertices.put(nodeId, v);
                scope.addVertex(v);
            }
            s.addVertex(v);

            // add as child of parent

            s.children.clear();
            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }
        }

        public void enterNode_id( DOTParser.Node_idContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterNode_id() ... \n");

            // add partial state
            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterNode_id() ... \n");
        }

        public void exitNode_id( DOTParser.Node_idContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitNode_id() ... \n");

            if (stack.isEmpty()) {
                return;
            }

            // Collect only first child (ignore ports)

            State s = stack.pop();
            if (!s.children.isEmpty()) {
                s.addId(s.children.getFirst().getId());

                // Add as child of parent
                s.children.clear();
                if (!stack.isEmpty()) {
                    stack.element().children.addLast(s);
                }
            }

            // Find attributes (last child)

            System.out.printf("---     Find attributes (last child) ... \n");

            Map<String, Attribute> attrs = null;
            State last = s.children.peekLast();
            if (last != null && last.attrs != null) {
                attrs = last.attrs;
            }

            System.out.printf("---     TEST: ADD VERTEX TO GRAPH HERE: s.id = %s ... \n", s.getId() );

            String nodeId = s.getId();

            System.out.printf("---     nodeId = %s ... \n", nodeId );

            // Create or update vertex

            V v = vertices.get(nodeId);
            if (v == null) {

                System.out.printf("---     Creating new vertex ... \n");

                // Find and append default attributes

                System.out.printf("---     Find and append default attributes ... \n");

                SubgraphScope scope = subgraphScopes.element();
                Map<String, Attribute> defaultAttrs = new HashMap<>(scope.nodeAttrs);
                if(attrs != null)
                  defaultAttrs.putAll(attrs);

                System.out.printf("---     Call vertexProvider ... \n");

                try {
                    v = vertexProvider.buildVertex(nodeId, defaultAttrs);
                } catch (Exception e) {
                    throw new IllegalArgumentException( "Vertex creation failed: " + e.getMessage(), e);
                }

                // Add vertex to graph ...

                System.out.printf("---     Run graph.addVertex(v) ... \n");

                graph.addVertex(v);

                // Save vertex to node hashmap ...

                System.out.printf("---     Add v to HashMap ... \n");

                vertices.put(nodeId, v);

                // Add vertex to scope ...

                System.out.printf("---     Add v to scope ... \n");

                scope.addVertex(v);
            } 

            // Add vertex to stack of states ...

            System.out.printf("---     Add v to state %s ... \n", s.getId() );

            s.addVertex(v);

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitNode_id() ... \n");
        }

        public void enterSubgraph( DOTParser.SubgraphContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterSubgraph() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            // Create new scope with inherited attributes

            Map<String, Attribute> defaultGraphAttrs = subgraphScopes.element().graphAttrs;
            Map<String, Attribute> defaultNodeAttrs  = subgraphScopes.element().nodeAttrs;
            Map<String, Attribute> defaultEdgeAttrs  = subgraphScopes.element().edgeAttrs;
            SubgraphScope newState = new SubgraphScope();
            newState.graphAttrs.putAll(defaultGraphAttrs);
            newState.nodeAttrs.putAll(defaultNodeAttrs);
            newState.edgeAttrs.putAll(defaultEdgeAttrs);
            subgraphScopes.push(newState);

            // Add partial state

            State s = new State();
            s.subgraph = newState;
            stack.push(s);

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterSubgraph() ... \n");
        }

        public void exitSubgraph( DOTParser.SubgraphContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Exit DOTImporter.CreateGraphDOTListener.exitSubgraph() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty() || subgraphScopes.isEmpty()) {
                return;
            }

            // Remove last scope

            SubgraphScope scope = subgraphScopes.pop();
            State s = stack.pop();

            // If not on root graph, append nodes to subgraph one level up

            if (scope.vertices != null && subgraphScopes.size() > 1) {
                subgraphScopes.element().addVertices(scope.vertices);
            }

            // Add as child of parent

            s.children.clear();
            if (!stack.isEmpty()) {
                stack.element().children.addLast(s);
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitSubgraph() ... \n");
        }

        // =====================================================================
        // ID Pairs are used in parsing of html ....
        // =====================================================================

        public void enterIdPair( DOTParser.IdContext ctx) {

            stack.push(new State());
        }

        public void exitIdPair( DOTParser.IdContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitIdPair() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty()) {
                return;
            }

            // Collect our two children as one pair

            State s = stack.pop();
            Iterator<State> it = s.children.iterator();
            if (it.hasNext()) {
                s.addId(it.next().getId());
            }
            if (it.hasNext()) {
                s.addId(it.next().getId());
            }

            if (s.ids != null) {
                // add as child of parent
                s.children.clear();
                if (!stack.isEmpty()) {
                    stack.element().children.addLast(s);
                }
            }

            System.out.printf("---     Exit DOTImporter.CreateGraphDOTListener.exitIdPair() ... \n");
        }

        public void enterId( DOTParser.IdContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.enterId() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            stack.push(new State());

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.enterId() ... \n");
        }

        public void exitId( DOTParser.IdContext ctx) {

            System.out.printf("--- \n");
            System.out.printf("---     Enter DOTImporter.CreateGraphDOTListener.exitId() ... \n");
            System.out.printf("---     Context ctx: %s ... \n", ctx.getText() );

            if (stack.isEmpty()) {
                return;
            }

            // Collect actual identifier

            System.out.printf("---     Collect identifier ... \n");

            State s = stack.pop();
            String id = null;
            if (ctx.ID() != null) {
                id = ctx.ID().toString();
                System.out.printf("---     id = %s ... \n", id );
            } else if (ctx.STRING() != null) {
                id = unescapeId(ctx.STRING().toString());
                System.out.printf("---     id = %s ... \n", id );
            } else if (ctx.HTML_STRING() != null) {
                id = unescapeHtmlString(ctx.HTML_STRING().toString());
                System.out.printf("---     id = %s ... \n", id );
            } else if (ctx.NUMBER() != null) {
                id = ctx.NUMBER().toString();
                System.out.printf("---     id = %s ... \n", id );
            }

            // Print id ...

            if (id != null) 
               System.out.printf("---     Save: id = %s to state ... \n", id );
            else
               System.out.printf("---     Save: failed ... \n" );

            // Record id ...

            if (id != null) {

                s.addId(id);

                // Add as child of parent

                if (!stack.isEmpty()) {
                    stack.element().children.addLast(s);
                }
            }

            System.out.printf("---     Leave DOTImporter.CreateGraphDOTListener.exitId() ... \n");
        }
    }

    /*
     * Partial parsed state depending on node type.
     */

    private class State {

        LinkedList<State> children;
        List<String> ids;
        Map<String, Attribute> attrs;
        List<V> vertices;
        SubgraphScope subgraph;

        public State() {
            this.children = new LinkedList<>();
            this.ids   = null;
            this.attrs = null;
            this.vertices = null;
            this.subgraph = null;
        }

        public String getId() {
            if (ids.isEmpty()) {
                return "null";
            } else {
                return ids.get(0);
            }
        }

        public void addId(String id) {
            if (this.ids == null) {
                this.ids = new ArrayList<>();
            }
            this.ids.add(id);
        }

        public void put(String key, Attribute value)
        {
            if (this.attrs == null) {
                this.attrs = new HashMap<>();
            }
            this.attrs.put(key, value);
        }

        public void putAll(Map<String, Attribute> attrs) {
            if (this.attrs == null) {
                this.attrs = new HashMap<>();
            }
            this.attrs.putAll(attrs);
        }

        public void addVertex(V v) {
            if (this.vertices == null) {
                this.vertices = new ArrayList<>();
            }
            this.vertices.add(v);
        }

        public List<V> getVertices() {
            if (vertices != null) {
                return vertices;
            } else if (subgraph != null && subgraph.vertices != null) {
                return subgraph.vertices;
            }
            return Collections.emptyList();
        }
    }

    /*
     * Records default attributes per subgraph
     */

    private class SubgraphScope {

        Map<String, Attribute> graphAttrs;
        Map<String, Attribute> nodeAttrs;
        Map<String, Attribute> edgeAttrs;
        List<V> vertices;

        public SubgraphScope() {

            this.graphAttrs = new HashMap<>();
            this.nodeAttrs = new HashMap<>();
            this.edgeAttrs = new HashMap<>();
            this.vertices = null;
        }

        public void addVertex(V v)
        {
            if (this.vertices == null) {
                this.vertices = new ArrayList<>();
            }
            this.vertices.add(v);
        }

        public void addVertices(List<V> v)
        {
            if (this.vertices == null) {
                this.vertices = new ArrayList<>();
            }
            this.vertices.addAll(v);
        }
    }

    /**
     * Unescape a string DOT identifier.
     *
     * @param input the input
     * @return the unescaped output
     */
    private static String unescapeId(String input)
    {
        final char QUOTE = '"';
        if (input.charAt(0) != QUOTE || input.charAt(input.length() - 1) != QUOTE) {
            return input;
        }
        String noQuotes = input.subSequence(1, input.length() - 1).toString();
        String unescaped = UNESCAPE_ID.translate(noQuotes);
        return unescaped;
    }

    /**
     * Unescape an HTML string DOT identifier.
     *
     * @param input the input
     * @return the unescaped output
     */

    private static String unescapeHtmlString(String input) {
        if (input.charAt(0) != '<' || input.charAt(input.length() - 1) != '>') {
            return input;
        }
        String noQuotes = input.subSequence(1, input.length() - 1).toString();
        String unescaped = StringEscapeUtils.unescapeXml(noQuotes);
        return unescaped;
    }
}
